﻿'Custom GUI's allow you to extend the look and behavior of blocks.
'
'This file implements the base class for custom GUI's. 
'You can interact with the block's parameters and the variables in end user
'12Block program.  You react to mouse events by first tracking mouse moves to set your "mode" and then using the "mode" to
'handle the click or drag.  Remember to call the "gui" class handlers for resizing...  12Blocks will set
'the "values" arrays with your variable values prior to calling the "draw" function.
Imports System.Windows.Forms, System.Drawing, System.IO.File
Public Class gui    
    Public rect As Rectangle = New Rectangle(0, 0, 10, 10) 'this blocks area on screen
    Public name As String  'name of block, corresponds to xml file
    Public basedir As String 'directory where image files are kept
    Public mode As Integer 'used for clicking/dragging
    Public values As Dictionary(Of String, Object) = New Dictionary(Of String, Object) 'set by 12Blocks to indicate variable values
    Public setpoints() As Single 'configuration points set by xml
    Event mouseevt(ByVal s As Object, ByVal e As MouseEventArgs) 'raise mouse event to 12b
    Event setVarE(ByVal var As String, ByVal value As Object) 'raise setting a variable
    Event setParamE(ByVal var As String, ByVal value As String) 'raise setting a parameter
    Event enterTextE(ByVal px As Integer, ByVal py As Integer, ByVal pm As String) 'start entering text for a parameter
    Event needControlE() 'gui needs control to be made
    Public ctrl As Control
    Public Overridable Sub makeControl()  'control is made
    End Sub
    Public Overridable Sub rmControl() 'control removed
    End Sub
    Public myFont As Font
    Public sprites As spriteC = New spriteC
    Public myblock As Object
    Protected myBrush As Brush
    Public mySheet As Control
    ''' <summary>
    ''' Called at compile time to get possibly dynamically generated code for this block
    Public Overridable Function getCode() As Xml.XmlElement
        Return Nothing
    End Function
    ''' <summary>
    ''' Called when cgui is initialized, right before parameters set and activated
    Public Overridable Sub init()
    End Sub
    ''' <summary>
    ''' Call to register a wizard to display a parameter region to display text and bring up a contol when clicked


    ''' <summary>
    ''' Call to set a program variable
    ''' </summary>
    ''' <param name="name">Name of variable to set</param>
    ''' <param name="value">Value of variable being set</param>
    Protected Sub setVar(ByVal var As String, ByVal value As Object)
        RaiseEvent setVarE(var, value)
    End Sub
    ''' <summary>
    ''' Called to clear blocks state
    ''' </summary>
    Public Overridable Sub clear()
    End Sub
    ''' <summary>
    ''' Call to raise a mouse event
    ''' </summary>
    Protected Sub setMouseEvent(ByVal sender As Object, ByVal e As MouseEventArgs)
        RaiseEvent mouseevt(sender, e)
    End Sub
    ''' <summary>
    ''' Call to set parameter of the block
    ''' </summary>
    ''' <param name="name">Name of parameter to set</param>
    ''' <param name="value">Value of parameter being set</param>
    Public Sub setparam(ByVal name As String, ByVal value As String)
        RaiseEvent setParamE(name, value)
    End Sub
    Public Sub needControl()
        RaiseEvent needControlE()
    End Sub
    ''' <summary>
    ''' Call to enter text with a textbox at a position for a named parameter
    ''' </summary>
    ''' <param name="px">X position of textbox</param>
    ''' <param name="py">Y position of textbox</param>
    ''' <param name="pm">Name of parameter to set</param>
    Protected Sub enterText(ByVal px As Integer, ByVal py As Integer, ByVal pm As String)
        RaiseEvent enterTextE(px, py, pm)
    End Sub
    ''' <summary>
    ''' Called to set parameters of the block as defined by xml
    ''' </summary>
    ''' <param name="name">Name of parameter being set</param>
    ''' <param name="value">Value of parameter being set</param>
    Public Overridable Sub updateParameter(ByVal name As String, ByVal value As Object)
        Dim n As Integer
        Select Case name
            Case "font" : myFont = CType(value, Font)
            Case "brush" : myBrush = CType(value, Brush)
            Case "size" : rect.Width = str2int(value, 0) : rect.Height = str2int(value, 1) : sprites.update(rect)
            Case "position" : rect.X = str2int(value, 0) : rect.Y = str2int(value, 1)
            Case "setpoints"
                Dim sp() As String = CStr(value).Split(","c)
                For n = 0 To Math.Min(sp.Length, setpoints.Length) - 1
                    Dim a As String = sp(n)
                    If a.StartsWith(".") Then a = "0" + a
                    setpoints(n) = Val(a)
                Next
            Case "look" : sprites.update(basedir, rect, value)
        End Select
    End Sub
    ''' <summary>
    ''' Called when Block has a new configuration
    ''' </summary>    
    Public Overridable Sub cfg()
    End Sub
    ''' <summary>
    ''' Called when any windows controls visible must change
    ''' </summary>    
    Public Overridable Sub setvisible(ByVal b As Boolean)
    End Sub
    ''' <summary>
    ''' Called to Draw the static part of this block.  values array and rect are set prior to call
    ''' </summary>
    ''' <param name="g">The graphic to draw to.</param>
    Public Overridable Sub drawStatic(ByVal g As Graphics)
        drawDynamic(g)
    End Sub
    ''' <summary>
    ''' Called to Draw the dynamic part of this block to animate it.  values array and rect are set prior to call
    ''' </summary>
    ''' <param name="g">The graphic to draw to.</param>
    Public Overridable Function drawDynamic(ByVal g As Graphics) As Boolean
    End Function
    ''' <summary>
    ''' Called to Ask if this block needs to be drawn. values array and rect are set prior to call
    ''' </summary>
    Public Overridable Function updateDynamic() As Boolean
        Return False
    End Function
    ''' <summary>
    ''' Called when 12Blocks changes state
    ''' </summary>
    ''' <param name="s">The new state:initialize,finalize,run,stop</param>
    Public Overridable Sub changeState(ByVal s As String)
        Debug.Print(s)
    End Sub

    ''' <summary>
    ''' Called when this block is clicked
    ''' </summary>
    ''' <param name="p">The position.</param>
    Public Overridable Sub OnClick(ByVal p As Point)
    End Sub
    ''' <summary>
    ''' Reacts to a right mouse click to edit context menu and position prior to show
    ''' </summary>
    ''' <param name="menu">The menu to show</param>
    ''' <param name="p">The position to show menu at.</param>
    Public Overridable Sub showContextMenu(ByVal c As Control, ByRef menu As ContextMenuStrip, ByRef p As Point)
        menu.Show(c, p)
    End Sub
    ''' <summary>
    ''' Reacts to a mouse drag
    ''' </summary>
    ''' <param name="start">The position where drag started.</param>
    ''' <param name="curp">The current position.</param>
    ''' <returns>Cursor to use</returns>
    Public Overridable Function onDrag(ByVal start As Point, ByVal current As Point) As Cursor
        Console.Write("dr")
        If mode = -1 Then
            Me.setparam("size", String.Format("{0},{1}", current.X - rect.X, current.Y - rect.Y))
            Return Cursors.SizeNWSE
        End If
        Return Cursors.Default
    End Function
    ''' <summary>
    ''' Reacts to a mouse move
    ''' </summary>
    ''' <param name="p">The position of the mouse</param>
    ''' <returns>A moveres which contains the cursor to use and mode to set to</returns>
    Public Overridable Function onMove(ByVal p As Point) As moveres
        If near(p, rect.Right, rect.Bottom) Then
            Return New moveres(Cursors.SizeNWSE, -1)
        End If

        Return New moveres(Cursors.Hand, 0)
    End Function
    'Helper functions
    Protected Function str2int(ByVal s As Object, Optional ByVal idx As Integer = 0) As Integer
        If idx = 0 Then Return CInt(Val(CStr(s)))
        Return CInt(Val(CStr(s).Split(","c)(idx)))
    End Function
    Protected Function near(ByVal p As Point, ByVal sx As Integer, ByVal sy As Integer) As Boolean
        Return Math.Abs(p.X - sx) + Math.Abs(p.Y - sy) < 10
    End Function
    Public Overridable Sub activate()
    End Sub
    Public Overridable Sub deactivate()
    End Sub
End Class


Public Class wizardgui
    Inherits gui
    Protected xoff, yoff As Integer
    Protected rct As Rectangle = New Rectangle

    Dim paramL As New List(Of paramI)
    Public Delegate Sub doClick(ByVal p As Point, ByVal f As Integer)
    Public Delegate Sub gotText(ByVal name As String, ByVal value As String)
    Public title As String
    Protected titleParam As paramI
    Dim gotTextF As gotText
    Dim tb As TextBox

    Public Sub New()
        titleParam = createParam(-1, "", AddressOf showTitleTB)
    End Sub
    Protected Sub clearParams()
        paramL.Clear()
    End Sub
    Private Sub showTitleTB(ByVal p As Point, ByVal f As Integer)
        showTextBox(AddressOf setTitle, titleParam.r.X, titleParam.r.Y, title, "Set title")
    End Sub
    Private Sub setTitle(ByVal n As String, ByVal s As String)
        title = s
        setparam("title", title)
    End Sub

    Public Sub showTextBox(ByVal _gotTextF As gotText, ByVal tx As Integer, ByVal ty As Integer, ByVal tname As String, ByVal tvalue As String)
        tb = New TextBox
        gotTextF = _gotTextF
        With tb
            .Top = ty
            .Name = tvalue
            .Text = ""
            .Left = tx
            .Width = rct.Right - tx - 1
            .Text = tname
            .BackColor = Color.Yellow
            mySheet.Controls.Add(tb)
            .SelectAll()
            .Visible = True
            .Select()
            AddHandler tb.Leave, AddressOf textdone
        End With
    End Sub
    Public Sub textdone(ByVal sender As Object, ByVal e As System.EventArgs)
        gotTextF(tb.Name, tb.Text)
        RemoveHandler tb.Leave, AddressOf textdone
        mySheet.Controls.Remove(tb)
    End Sub

    Public Class paramI
        Public xoff, yoff, wid, nameof, firsts As Integer 'first set point
        Public name, value As String 'name,value to show
        Public f As doClick 'function to call
        Public r As New Rectangle 'parameter rectangle
        Public p As gui
        Public Sub New(ByVal par As gui, ByVal _s As Integer, ByVal _nm As String, ByVal _f As doClick)
            name = _nm : f = _f : firsts = _s
            p = par
        End Sub
    End Class
    'open a control, or call a function on click
    Public Overridable Function createParam(ByVal firstSetPt As Integer, ByVal name As String, ByVal f As doClick) As paramI
        'wizard(first of 2 setpoints, text, control to show,function to call)
        'wizards use modes>1000
        Dim w As New paramI(Me, firstSetPt, name, f)
        paramL.Add(w)
        Return w
    End Function
    Public Overrides Function onMove(ByVal p As System.Drawing.Point) As GUIBase.moveres
        Dim n As Integer = 1000
        For Each w As paramI In paramL
            If w.r.Contains(p) Then Return New GUIBase.moveres(Cursors.IBeam, n)
            n += 1
        Next
        Return MyBase.onMove(p)
    End Function

    Public Overrides Sub OnClick(ByVal p As Point)
        If mode >= 1000 And mode - 1000 < paramL.Count Then
            With paramL(mode - 1000)
                .f(p, .firsts)
            End With
        Else
            MyBase.OnClick(p)
        End If
        Return
    End Sub
    Public Overrides Sub drawstatic(ByVal g As System.Drawing.Graphics)
        'fills entire rectangle and borders
        rct.X = rect.X : rct.Y = rect.Y : rct.Width = rect.Width : rct.Height = rect.Height
        g.FillRectangle(Brushes.White, rct)
        g.DrawRectangle(Pens.Black, rct)
        'sets rct to xoff/yoff
        rct.Y += 1 + yoff : rct.Height -= 1 + yoff
        rct.X += 1 + xoff : rct.Width -= 1 + xoff
        titleParam.value = title
        'draws parameters
        For Each w As paramI In paramL
            With w
                If .firsts = -1 Then
                    .xoff = 1
                    .yoff = 1
                    .wid = -1
                    .nameof = -1
                Else
                    If .p.setpoints Is Nothing Then Return
                    .xoff = CInt(.p.setpoints(.firsts + 0))
                    .yoff = CInt(.p.setpoints(.firsts + 1))
                    .wid = CInt(.p.setpoints(.firsts + 2))
                    .nameof = CInt(.p.setpoints(.firsts + 3))
                End If
                .r.Height = .p.myFont.Height
                .r.X = rect.X + .xoff
                .r.Y = rect.Y + .yoff
                .r.Width = If(.wid = -1, rect.Right - .r.X, .wid)
                g.FillRectangle(Brushes.Yellow, .r)
                g.DrawString(.value, myFont, Brushes.Black, .r.X, .r.Y)
                g.DrawString(.name, myFont, Brushes.Black, .r.X + .nameof, .r.Y)
            End With
        Next
    End Sub

   
End Class

'sprites maintain either an image or color for drawing by blocks
'they're specified by a "look" value and "ratio" and are updated when their size changes.
Public Class spriteC
    Public imgS() As Image
    Public origImgS() As Image
    Public fileS() As String
    Public colorS() As Color
    Public ratioS() As SizeF
    Public Sub init(ByVal num As Integer)

        Dim n, nalloc As Integer
        nalloc = num - 1
        ReDim imgS(nalloc), origImgS(nalloc), colorS(nalloc), ratioS(nalloc), fileS(nalloc)
        For n = 0 To num - 1
            colorS(n) = Color.Black
            ratioS(n) = New Size(1, 1)
        Next
    End Sub
    Public Sub fillRectangle(ByVal g As Graphics, ByVal idx As Integer, ByVal r As Rectangle)
        If imgS(idx) IsNot Nothing Then
            g.DrawImage(imgS(idx), r.X, r.Y)
        Else
            g.FillRectangle(New SolidBrush(colorS(idx)), r)
        End If
    End Sub
    Public Sub update(ByVal r As Rectangle)
        If r.Width <= 0 Or r.Height <= 0 Then Return
        Dim n As Integer
        Dim i As Image

        For n = 0 To origImgS.Length - 1
            If ratioS(n).Width = 0 Then ratioS(n).Width = 1 : ratioS(n).Height = 1
            If origImgS(n) IsNot Nothing Then
                i = New Bitmap(Math.Max(CInt(r.Width * ratioS(n).Width), 1), Math.Max(CInt(r.Height * ratioS(n).Height), 1))
                Graphics.FromImage(i).DrawImage(origImgS(n), 0, 0, i.Width + 1, i.Height + 1)
                imgS(n) = i
            Else
                imgS(n) = Nothing
            End If
        Next
    End Sub
    Public Sub update(ByVal basedir As String, ByVal r As Rectangle, ByVal l As String)
        Dim look As String() = l.Split(","c)
        Dim n, cv As Integer
        Dim c As Color
        For n = 0 To Math.Min(look.Length - 1, colorS.Length - 1)
            c = Color.FromName(look(n))
            fileS(n) = ""
            If c.A = 255 Then
                colorS(n) = c
            ElseIf Exists(look(n)) Then
                fileS(n) = look(n)
            ElseIf Exists(basedir & look(n)) Then
                fileS(n) = basedir & look(n)
            Else
                Try
                    cv = Integer.Parse(look(n), Globalization.NumberStyles.HexNumber)
                    colorS(n) = Color.FromArgb(cv.ToString)
                Catch ex As Exception
                    Debug.Print("Bad look:" & look(n))
                    colorS(n) = Color.Black
                End Try
            End If
            origImgS(n) = If(fileS(n) > "", Bitmap.FromFile(fileS(n)), Nothing)
        Next
        update(r)
    End Sub
End Class

Public Class moveres
    ''' <summary>
    ''' The cursor to show
    ''' </summary>
    Public curs As Cursor
    ''' <summary>
    ''' The mode to use for follow-on clicks/drags
    ''' </summary>
    Public mode As Integer
    Public Sub New(ByVal _cursor As Cursor, ByVal _mode As Integer)
        curs = _cursor : mode = _mode
    End Sub
End Class

