﻿using System;
using System.Collections.Generic;
using System.Text;

using System.Windows.Forms;
using System.Drawing;

using GUIBase;
using Microsoft.DirectX.DirectInput;

namespace GuiBlocks
{
   
    public static class JoystickManager
    {
        public static List<Device> Joysticks = new List<Device>();
 
        static JoystickManager()
        {
            DeviceList gameControllerList = Manager.GetDevices(DeviceClass.GameControl,
                        EnumDevicesFlags.AttachedOnly);

            for (int i = 0; i < gameControllerList.Count; i++)
            {
                gameControllerList.MoveNext();
                DeviceInstance di = (DeviceInstance)gameControllerList.Current;

                Device d = new Device(di.InstanceGuid);
                d.SetCooperativeLevel(null, CooperativeLevelFlags.Background | CooperativeLevelFlags.NonExclusive);
                d.SetDataFormat(DeviceDataFormat.Joystick);
                d.Acquire();
                Joysticks.Add(d);
            }
        }

    }
   
    public class gamepadGui : gui
    {
        
        string varX1;
        string varY1;
        string varX2;
        string varY2;
        string varBtn;
        string varPov;

        int port;
        int min_axis;
        int max_axis;
        int[] gvalues;

        Font labelFont = new Font("Arial", 8);
        SolidBrush labelBrush = new SolidBrush(Color.Blue);

        public gamepadGui()
            : base()
        {
            gvalues = new int[6];
            setpoints=new float[4];
            sprites.init(1);
            name = "gamepad";

        }
        
        private Rectangle GetImageRect()
        {
            Image img = Resource1.LogitechGamepad;
            
            float scaleX = rect.Width / (float)img.Width;
            float scaleY = rect.Height / (float)img.Height;

            float scale = scaleX > scaleY ? scaleY : scaleX;

            Rectangle r = new Rectangle(rect.X, rect.Y, (int)(img.Width * scale), (int)(img.Height * scale));
            return r;
        }
        public override moveres onMove(Point p)
        {
            Rectangle r = GetImageRect();

            if (near(p, r.Right, r.Bottom))
                return new moveres(Cursors.SizeNWSE, -1);

            float xpos = (p.X - r.X) / (float)r.Width;
            float ypos = (p.Y - r.Y) / (float)r.Height;

            //Console.WriteLine("{0}, {1}", xpos, ypos);

            return new moveres(Cursors.Hand, 0);
        }
        
        public override Cursor onDrag(Point start, Point current)
        {
             if (mode == -1)
             {
                setparam("width", (current.X - rect.X).ToString());
                setparam("height", (current.Y - rect.Y).ToString());

                return Cursors.SizeNWSE;
             }
            
            return Cursors.Default;
        }

        private PointF GetLabelPos(Rectangle limits, int index)
        {
            float[] pts = new float[] { 0.500F, 0.900F,     // Connection 
                                        0.345F, 0.610F,     // Left Thumbstick
                                        0.648F, 0.610F,     // Right Thumbstick
                                        0.199F, 0.373F,     // Top Hat

                                        0.500F, 0.500F,     // Not Used
                            
                                        0.718F, 0.374F,     // Button #1
                                        0.796F, 0.494F,     // Button #2
                                        0.875F, 0.374F,     // Button #3
                                        0.796F, 0.265F,     // Button #4

                                        0.199F, 0.120F,    // Button #5,#7
                                        0.794F, 0.120F,    // Button #6,#8
                                        0.199F, 0.120F,    // Button #5,#7
                                        0.794F, 0.120F,    // Button #6,#8
                                        
                                        0.392F, 0.273F,     // Button #9
                                        0.605F, 0.273F,     // Button #10
                            };

            float xval = (pts[index * 2] * limits.Width + limits.X);
            float yval = (pts[index * 2+1] * limits.Height + limits.Y);
            return new PointF(xval, yval);
        }

        private void drawLabel(Graphics g, string txt, int index, Rectangle limits, bool border)
        {
            drawLabel(g, txt, index, limits, labelFont, labelBrush, border);
        }

        private void drawLabel(Graphics g, string txt, int index, Rectangle limits, Font font, Brush brush, bool border)
        {
            PointF pt;
            SizeF sz;

            pt = GetLabelPos(limits, index);
            sz = g.MeasureString(txt, font);

            if (border)
            {
                Rectangle r = new Rectangle((int)(pt.X - sz.Width / 2.0F), (int)(pt.Y - sz.Height / 2.0F), (int)sz.Width, (int)sz.Height);
                r.Inflate(2, 2);
                g.FillRectangle(Brushes.BlanchedAlmond, r);
                g.DrawRectangle(Pens.Black, r);
            }

            pt.X -= sz.Width / 2.0F;
            pt.Y -= sz.Height / 2.0F;

            g.DrawString(txt, font, brush, pt);
        }

        
        public override bool  drawDynamic(System.Drawing.Graphics g)
        {
            bool valid = updateJoystick();

            Rectangle r = GetImageRect();
            g.DrawImage(Resource1.LogitechGamepad, r);

            
            // Port Connection
            if (valid)
            {
                drawLabel(g, String.Format("Controller {0}", port + 1), 0, r, labelFont, Brushes.Black, false);
            }
            else
            {
                drawLabel(g, String.Format("Not Connected", port + 1), 0, r,labelFont, Brushes.Red, false);
                return false;
            }

            
            // Left and Right Thumbstick
            for (int i = 0; i < 2; i++)
            {
                drawLabel(g, String.Format("{0},{1}", gvalues[i * 2], gvalues[i * 2 + 1]), 
                    i + 1, r, true);
            }

            // Top Hat
            drawLabel(g, gvalues[5].ToString(), 3, r,true);

            // Buttons
            //drawLabel(g, values[4].ToString(), 4, r);

            for (int i = 0; i < 10; i++)
            {
                //PointF pt = GetLabelPos(r, 5 + i);
                if ((gvalues[4] & (1 << i)) != 0)
                {
                    drawLabel(g, gvalues[4].ToString(), 5 + i, r, true);
                }
            }
            return false;
        }

        public bool updateJoystick()
        {
            
            try
            {
                // poll the joystick
                Device d = JoystickManager.Joysticks[port];
                d.Poll();
                
                JoystickState state = d.CurrentJoystickState;
               
                gvalues[0] = state.X;
                gvalues[1] = state.Y;
                gvalues[2] = state.Z;
                gvalues[3] = state.Rz;

                for (int i = 0; i < 4; i++)
                {
                    gvalues[i] = (int)((((max_axis - min_axis) / 65536.0) * gvalues[i]) + min_axis);
                }

                byte[] btns = state.GetButtons();
                int btn_val = 0;
                for (int i = 0; i < 32; i++)
                {
                    if (btns[i] >= 128)
                    {
                        btn_val += 1 << i;
                    }
                }

                gvalues[4] = btn_val;
                
                int[] pov = state.GetPointOfView();
                gvalues[5] = pov[0];
                if (gvalues[5] > 0)
                    gvalues[5] /= 100;

                setVar(varX1, gvalues[0].ToString());
                setVar(varY1, gvalues[1].ToString());
                setVar(varX2, gvalues[2].ToString());
                setVar(varY2, gvalues[3].ToString());
                setVar(varBtn, gvalues[4].ToString());
                setVar(varPov, gvalues[5].ToString());

                //Console.WriteLine("X1,Y1: {0}, {1}   X2,Y2: {2}, {3}   Btn: {4}   Pov: {5}",
                //                values[0], values[1], values[2], values[3], values[4], values[5]);
                return true;
            }
            catch (Exception )
            {
                // Console.WriteLine(err.Message);
                return false;
            }
             

        }

        public override void updateParameter(string item, object value)
        {
            switch (item)
            {
                case "port":
                    int.TryParse((string)value, out port);
                    port -= 1;
                    break;
                case "min":
                    int.TryParse((string)value, out min_axis);
                    break;
                case "max":
                    int.TryParse((string)value, out max_axis);
                    break;

                case "varx1":
                    varX1 = (string)value;
                    break;
                case "vary1":
                    varY1 = (string)value;
                    break;
                case "varx2":
                    varX2 = (string)value;
                    break;
                case "vary2":
                    varY2 = (string)value;
                    break;
                case "varbtn":
                    varBtn = (string)value;
                    break;
                case "varpov":
                    varPov = (string)value;
                    break;
                default:
                    base.updateParameter(item, value);
                    break;
            }
        }
   


    }

}
