#include <arduino.h>
#include "or_arm.h"
#define SPI_MODE_MASK 0x0C  // CPOL = bit 3, CPHA = bit 2 on SPCR
#define SPI_CLOCK_MASK 0x03  // SPR1 = bit 1, SPR0 = bit 0 on SPCR
#define SPI_2XCLOCK_MASK 0x01  // SPI2X = bit 0 on SPSR
#define SPI_CLOCK_DIV2 0x04

void _OR_arm::begin() {
	unsigned char n;	
	for(n=0;n<_SERVOS;n++) {_servo_pos[n]=255;}
	for(n=10;n<13;n++) digitalWrite(n,LOW);	
	_latch=13; //will get ++
	_count=254;	//causes immediated calculation of events
	_pos=255;
	//init spdr
	digitalWrite(SCK, LOW);
	digitalWrite(MOSI, LOW);
	digitalWrite(SS, HIGH);
	SPCR |= _BV(MSTR);
	SPCR |= _BV(SPE);
	int rate=SPI_CLOCK_DIV2 ;
	SPCR = (SPCR & ~SPI_CLOCK_MASK) | (rate & SPI_CLOCK_MASK);
	SPSR = (SPSR & ~SPI_2XCLOCK_MASK) | ((rate >> 2) & SPI_2XCLOCK_MASK);
}
int _OR_arm::sense(char x) {
//0-5:feedback,6-mic,7=btn,8=ping
	switch(x) {
		case 0: //line done by isr
		case 1:
		case 2:
		case 3:return analogRead(A0+x);
		case 4:return analogRead(A5);
		case 5:return analogRead(A7);
		case 6:return analogRead(A6)-512; //mic
		case 7:return (1-digitalRead(D8))*2+(1-digitalRead(D9)); //btn
		case 8: //ping done by isr
			if(!doping) doping=1;
			return ping;		
	}
}		
void _OR_arm::isr(){
  unsigned char output,n,min,idx;  
  _count++;
  switch(doping) {
		case 1: //send trigger
			digitalWrite(2,0);		
			pinMode(2,INPUT);
			doping++;
			count=0;
		case 2: //wait for high
			if(digitalRead(3)||count++>200) {
				doping++;
				count=0;
			}
			break;
		case 3: //wait for low or timeout then store ping, and reset doping
			if(!digitalRead(3) || count++>200) {
				doping=0;
				ping=count;
			}			
  }
  if(_count==255) { //calculate next group, count is now 0
    if(++_latch<14) return; //need 4 groups 10,11,12,13
	_latch=10;
	_latchport=portOutputRegister(digitalPinToPort(_latch));
	_latchmask=digitalPinToBitMask(_latch);	
	
	if(_pos<255){
		_servo_cnt[0]=255;
		output=85; //odd servos on
	}
	else{
		//calculate events
		for(n=0;n<8;n++) _tmp[n]=_servo_pos[n]; //gets servo values for latch
		output=255; //start states as high, lower for each servo
		//
		
		for(_event=0;_event<_EVENTS;_event++) {
			min=_tmp[0];idx=0; //find least one
			for(n=1;n<8;n++)
				if(_tmp[n]<min) {min=_tmp[n];idx=n;}
			if(min==255) break;  //when minimum is 255 stop adding events
			if(_event>0 && _servo_cnt[_event-1]==min) _event--; //collapse events with same time
			_servo_cnt[_event]=min;
			//output &= ~(1<<idx); //finish servo's pulse '32107654
			output &= ~(1<<((11-idx)%8)); //finish servo's pulse
			_servo_out[_event]=output;
			_tmp[idx]=255; //don't process this servo again
		}
		_servo_cnt[_event]=255;	//last event is 255	
		output=255;//now finish by outputting start state	
	}	
	_event=0; //now set to first event			
  }
  else if(_count==_servo_cnt[_event]) {
    output=_servo_out[_event++]; //get pin states to output   	
  }
  else if(_count==_pos) {
	output=0; //all servos off
  }
  else
    return;  

  SPDR = output; // push the byte to be loaded to the SPI register
  while(!(SPSR & (1<<SPIF))); //wait till the register completes
  digitalWrite(_latch,LOW);
  digitalWrite(_latch,HIGH);
  //*_latchport &= ~_latchmask; // set register
  //*_latchport |= _latchmask;  // 
}
void _OR_arm::go(int s1,int s2,int s3,int s4,int s5){ //move servos
	_servo_pos[0]=s1;
	_servo_pos[1]=s2;
	_servo_pos[2]=s3;
	_servo_pos[3]=s4;
	_servo_pos[7]=s5;	
}
void _OR_arm::sweep(int s1,int s2,int s3,int s4,int s5,int wt){ //move servos over time
	unsigned int n,tmp[_SERVOS],_servo_tgt[_SERVOS];
	unsigned int fract,ifract;
	unsigned int m;	
	for(m=0;m<_SERVOS;m++)
		_servo_tgt[m]=255;
	_servo_tgt[0]=s1;
	_servo_tgt[1]=s2;
	_servo_tgt[2]=s3;
	_servo_tgt[3]=s4;
	_servo_tgt[7]=s5;	
	
	if(wt>100) {
		for(n=0;n<_SERVOS;n++) 
			if(_servo_pos[n]==255)
				tmp[n]=_servo_tgt[n];
			else
				tmp[n]=_servo_pos[n]; //copy current position
		for(m=0;m<wt;m+=10){  //adjust through time			
			fract = m/ (wt/100); //0..100. ok when wt>100
			ifract = 100 - fract; //100..0
			for(n=0;n<_SERVOS;n++) 
				_servo_pos[n]=((tmp[n]*ifract)+(fract*_servo_tgt[n]))/100;					
			pauset(10);				
		}		
	}	
	else
		pauset(wt);
	for(n=0;n<_SERVOS;n++) _servo_pos[n]=_servo_tgt[n]; 	
}