#include <arduino.h>
#include "or_scara.h"
void _OR_scara::begin() {
	_speed=3;
	int n,pin;
	//stp a0 a6 d46
	//dir a1 a7 d48
	//direction pins
	_dir[0]=55;//a1+54
	_dir[1]=61;//a7+54
	_dir[2]=48;
	//step pins
	_bit[0]=54;//a0+54
	_bit[1]=60;//a6+54
	_bit[2]=46;//
	
	_max[0]=3;
	_max[1]=14;
	_max[2]=18;
	for(n=0;n<_STEPN;n++){	
		pinMode(_max[n],INPUT_PULLUP);  
		pin=_bit[n];
		digitalWrite(pin,LOW);
		_bit[n] = digitalPinToBitMask(pin);
		_out[n] = portOutputRegister(digitalPinToPort(pin));		
		_next[n]=255;
	}			
	digitalWrite(38,0); //enable
	digitalWrite(A2,0);
	digitalWrite(A8,0);			
}
void _OR_scara::isr(){ 
	//event and next[] measure isrs until action happens
	//event=0..255,next 0..254 when active, 255 when not
	
	if(!(--_event)){
		uint8_t n;
		_event--;
		for(n=0;n<_STEPN;n++){
			if(!_next[n]) { 
					*_out[n] ^= _bit[n];
					_steps[n]--;   //take one step, calculate next step if any
					int _left=abs(abs(_steps[n])-_done[n]);
					uint8_t _lk[]={255,40,10,8,7,6,5,5,5,4,4,4,4,4,4};
					if(_left<15)					
						_next[n]=max(_lk[_left],_speed);					
					else
						_next[n]=max(3,_speed);					
			} 
			if(_next[n]<_event) //next event is closest
				_event=_next[n];
		}//loop steppers
		for(n=0;n<_STEPN;n++) //update all active events
			if(_next[n]<255) 
				_next[n]-=_event;		
	}
}
void _OR_scara::steps(int dx,int dy,int dz){	
	_steps[0]=dx;_steps[1]=dy;_steps[2]=-dz;
	steps();
}
void _OR_scara::steps(){	
	int n,_mm;
	_mm=0;
	for(n=0;n<_STEPN;n++) {	//write direction pin
		if(_steps[n]<0) {  //high when reversing, make steps positive
			digitalWrite(_dir[n],1);
			_steps[n]=-_steps[n];
			}
		else
			digitalWrite(_dir[n],0);		
		if(_steps[n]>=_steps[_mm])  //find longest one
			_mm=n;		
	}
	if(!_steps[_mm]) return;
	for(n=0;n<_STEPN;n++){	
		if(_steps[n]) {//activate non-0 steppers			
			_next[n]=0;
			_done[n]=_steps[n];
		}
	}	
	while(!(_next[0]==255 && _next[1]==255 && _next[2]==255)) {
		pauset(10);
	}	
}
void _OR_scara::getaxis(int x,int y,double *elbow,double *shoulder) {
	double ss1,ss2,ss,qa;  //calculate scara elbow/shoulder from cartesian model
	ss1=square((double)x) + square((double)y)-square(_l2);
	ss2=square(_l1);
	
	ss = (ss1-ss2) / (2.0 * _l1 * _l2);
	if(ss>1 || ss< -1)
		*elbow=0;
	else
		*elbow = acos(ss);
	ss = (ss1+ss2) / (2.0 * _l1 * sqrt(square((double)x) + square((double)y)));
	if(ss>1 or ss <-1)
		qa = 0;
	else
		qa=acos(ss);
	*shoulder = atan2(y, x)- qa;
}
void _OR_scara::move(int x,int y,int z){ //move by amount specified
	_x+=x;_y+=y;_z+=z;  //adjust cartesian model
	double elbow,shoulder;
	getaxis(_x,_y,&elbow,&shoulder);
		
	//calculate #steps to move by finding difference between desired shoulder/elbow and current sh/el
	_steps[0]=(shoulder*_l3)-_sh;
	_steps[1]=(elbow*_l4)-_el;
	_sh+=_steps[0];
	_el+=_steps[1];
	
	_steps[2]=-z;	
	steps();
}
void _OR_scara::target(int x,int y,int z){ //move to an exact cartesian
	move(x-_x,y-_y,z-_z);
}
void _OR_scara::speed(int x){ //100 is fastest
	_speed=constrain(103-x,3,100);
}
void _OR_scara::home(int hx,int hy,int hz){
	int n,m,nn;
	int ordr[]={2,0,1};
	for(nn=0;nn<_STEPN;nn++) {	//write direction pin
		n=ordr[nn];
		digitalWrite(_dir[n],1);
		digitalWrite(_dir[1],0);
		for(m=0;m<30000;m++){
			*_out[n] ^= _bit[n];
			delayMicroseconds(10);
			if(digitalRead(_max[n])) break;
		}
	}
	double elbow,shoulder;
	getaxis(hx,hy,&elbow,&shoulder);
	_sh=shoulder*_l3;
	_el=elbow*_l4;
	_x=hx;_y=hy;_z=hz;
	//uses x,y,z as current position
	
}
