''*******************
''*   MIDI player   *   without SD card
''*******************
'' Author:  Andy Schenk, Insonix
'' uses the PhaseMod GM Synthesizer to play a MIDI file.
'' The MIDI file must be not longer then 22 kByte to fit in the HubRAM.
'' Only the MidiFile Format 0 is supported (all Tracks mixed down to one).

'' Adapted for 12Blocks by Hanno Sander

CON
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000
VAR    
  long  time, tq, ts, us
  long  ramptr
OBJ
  synth : "Pm_synth_20.spin"
pub play(pntr,l,r)
  synth.start(l,r,2)
  midiplay(pntr)
  synth.alloff    

PUB midiplay(pntr)  | lg, dt, b, v, st, d1, d2, ch
    ramptr := pntr
    us := clkfreq / 1_000_000                           'ticks / microsecond
    b := st := d1 := d2 := v := 0
    pread(@v,4)
    if v == "M" + "T"<<8 + "h"<<16 + "d"<<24            'Header ID
      pread(@v,4)                                       'Header Len
      v~
      pread(@v,2)                                       'Format
      if v <> 0
'        disp.str(string("only Format 0 supported"))
      else
        pread(@v,2)                                     'Tracks
        pread(@v,2)                                     'Ticks/Quarter
        tq := v>>8 + (v&255)<<8
        ts := 500_000 / tq                              'default Tempo 120 BPM
        pread(@v,4)                                     '"MTrk"
        pread(@v,4)                                     'Track length
        lg := v>>24 + (v>>16&255)<<8 + (v>>8&255)<<16
        time := cnt
        repeat
          pread(@b,1)                                   'get delta time
          lg--
          dt := b & $7F
          repeat while b>127
            pread(@b,1)
            lg--
            dt := dt<<7 + (b & $7F)
          time := time + dt*us*ts
          repeat until time-cnt < 0                     'wait deltatime
          b~
          pread(@b,1)                                   'MIDI byte
          if b > 127
            st := b                                     'new status byte
          elseif st==$FF
            st~
          ch := st & $0F
          d1~
          d2~
          case st & $F0                                 'decode MIDI Event
            $90: pread(@d1,1)                           'Note 
                 pread(@d2,1)                           'Velocity
                 lg -= 3
                 if d2>0
                   synth.noteOn(d1,ch,d2)               'Note On
                   show(15,ch+1,string(127),-1)
                 else
                   synth.noteOff(d1,ch)                 'Note Off if Vel=0
                   show(15,ch+1,string("."),-1)
            $80: pread(@d1,1)                           'Note Off 
                 pread(@d2,1)                           'Velocity
                 lg -= 3
                 synth.noteOff(d1,ch)
                 show(15,ch+1,string("."),-1)
                 if ch==9
                   show(20,ch+1,string(">"),d1)
            $A0: pread(@d1,1)                           'Poly AfterTouch 
                 pread(@d2,1)                           '(not supported)
                 lg -= 3
            $C0: pread(@d1,1)                           'Prg Change 
                 lg -= 2
                 synth.prgChange(d1,ch)
                 show(0,ch+1,string("Ch:"),ch+1)
                 show(6,ch+1,string("Prg:"),d1)
            $D0: pread(@d1,1)                           'Mono AfterTouch 
                 lg -= 2                                '(not supported)
            $B0: pread(@d1,1)                           'Controller (nr) 
                 pread(@d2,1)                           'value
                 lg -= 3
                 if d1==7
                   synth.volContr(d2,ch)                '7=Vol
            $E0: pread(@d1,1)                           'Pitch Bender 
                 pread(@d2,1)                           '(not supported!)
                 lg -= 3
            $F0: pread(@d1,1)                           'Meta Event
                 pread(@b,1)                            'Len
                 lg -= 2
                 v~
                 repeat while b>0                       'read meta data
                   pread(@d2,1)
                   v := v<<8 + d2
                   b--
                   lg--
                   if d1==5 and v&$80 > 0
                     quit
                 if d1==81                              'Tempo
                   ts := v / tq
        until lg < 1
    else
'      disp.str(string("No MIDI file"))
    synth.allOff

PUB show(x,y,strp,val)                                  ''visualize MIDI events
PUB pread(rptr,size)                                    ''Read RAM data
  repeat size
    byte[rptr++] := byte[ramptr++]